@@ -18,16 +18,46 @@
 #include <immintrin.h>
 #endif
 
+#ifdef HAVE_AVX512
+/* Define __MM_MALLOC_H to prevent importing the memory aligned
+ * allocation functions, which we don't use. */
+#define __MM_MALLOC_H
+#include <immintrin.h>
+#endif
+
+#ifdef HAVE_AARCH64_NEON
+#include <arm_neon.h>
+#endif
+
 #ifdef HAVE_AVX2
 #define BITOP_USE_AVX2 (__builtin_cpu_supports("avx2"))
 #else
 #define BITOP_USE_AVX2 0
 #endif
 
+/* AArch64 NEON support is determined at compile time via HAVE_AARCH64_NEON */
+#ifdef HAVE_AVX512
+#define BITOP_USE_AVX512 (__builtin_cpu_supports("avx512f") && __builtin_cpu_supports("avx512vpopcntdq"))
+#else
+#define BITOP_USE_AVX512 0
+#endif
+
+
 /* -----------------------------------------------------------------------------
  * Helpers and low level bit functions.
  * -------------------------------------------------------------------------- */
 
+ /* Shared lookup table for bit counting - maps each byte value to its popcount */
+static const uint8_t bitsinbyte[256] = {
+    #define B2(n) n, n+1, n+1, n+2
+    #define B4(n) B2(n), B2(n+1), B2(n+1), B2(n+2)
+    #define B6(n) B4(n), B4(n+1), B4(n+1), B4(n+2)
+    B6(0), B6(1), B6(1), B6(2)
+    #undef B6
+    #undef B4
+    #undef B2
+};
+
 /* Count number of bits set in the binary array pointed by 's' and long
  * 'count' bytes. The implementation of this function is required to
  * work with an input string length up to 512 MB or more (server.proto_max_bulk_len) */
@@ -42,8 +72,6 @@ long long redisPopcount(void *s, long count) {
     int use_popcnt = 0; /* Assume CPU does not support POPCNT if
                          * __builtin_cpu_supports() is not available. */
 #endif
-    static const unsigned char bitsinbyte[256] = {0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4,1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,4,5,5,6,5,6,6,7,5,6,6,7,6,7,7,8};
-    
     /* Count initial bytes not aligned to 64-bit when using the POPCNT instruction,
      * otherwise align to 32-bit. */
     int align = use_popcnt ? 7 : 3;
@@ -121,6 +149,237 @@ remain:
     return bits;
 }
 
+#ifdef HAVE_AARCH64_NEON
+/* AArch64 optimized popcount implementation.
+ * Processes the input bitmap using four NEON vector accumulators in parallel
+ * to improve instruction-level parallelism and reduce the frequency of
+ * scalar reductions. Each accumulator holds 16-bit partial sums that are
+ * combined only once per large block (128 bytes), minimizing data movement.
+ *
+ * Benchmark results show this approach outperforms 2-lane implementations
+ * and matches or exceeds 8-lane versions in throughput, while avoiding
+ * register pressure and keeping the backend pipeline fully utilized.
+ *
+ * This function is now memory bound on large bitmaps, as confirmed by perf
+ * profiling, with backend stalls dominated by L1/L2 data cache refills.
+ */
+long long redisPopCountAarch64(void *s, long count) {
+    long long bits = 0;
+    const uint8_t *p = (const uint8_t*)s;
+
+    /* Align */
+    while (((uintptr_t)p & 15) && count) {
+        bits += bitsinbyte[*p++];
+        count--;
+    }
+
+    /* Four vector accumulators of u16 (pairwise-accumulated byte counts). */
+    uint16x8_t acc0 = vdupq_n_u16(0);
+    uint16x8_t acc1 = vdupq_n_u16(0);
+    uint16x8_t acc2 = vdupq_n_u16(0);
+    uint16x8_t acc3 = vdupq_n_u16(0);
+
+    /* Process 128B per loop to amortize reductions. */
+    while (count >= 128) {
+        uint8x16_t d0 = vld1q_u8(p +  0);
+        uint8x16_t d1 = vld1q_u8(p + 16);
+        uint8x16_t d2 = vld1q_u8(p + 32);
+        uint8x16_t d3 = vld1q_u8(p + 48);
+        uint8x16_t d4 = vld1q_u8(p + 64);
+        uint8x16_t d5 = vld1q_u8(p + 80);
+        uint8x16_t d6 = vld1q_u8(p + 96);
+        uint8x16_t d7 = vld1q_u8(p +112);
+
+        /* Per-byte popcount */
+        uint8x16_t c0 = vcntq_u8(d0);
+        uint8x16_t c1 = vcntq_u8(d1);
+        uint8x16_t c2 = vcntq_u8(d2);
+        uint8x16_t c3 = vcntq_u8(d3);
+        uint8x16_t c4 = vcntq_u8(d4);
+        uint8x16_t c5 = vcntq_u8(d5);
+        uint8x16_t c6 = vcntq_u8(d6);
+        uint8x16_t c7 = vcntq_u8(d7);
+
+        /* Pairwise widen-add with accumulation: u8 -> u16, stay in vectors */
+        acc0 = vpadalq_u8(acc0, c0);
+        acc1 = vpadalq_u8(acc1, c1);
+        acc2 = vpadalq_u8(acc2, c2);
+        acc3 = vpadalq_u8(acc3, c3);
+
+        acc0 = vpadalq_u8(acc0, c4);
+        acc1 = vpadalq_u8(acc1, c5);
+        acc2 = vpadalq_u8(acc2, c6);
+        acc3 = vpadalq_u8(acc3, c7);
+
+        p += 128;
+        count -= 128;
+    }
+
+    /* Reduce vector accumulators to scalar once. */
+    uint32x4_t s0 = vpaddlq_u16(acc0);
+    uint32x4_t s1 = vpaddlq_u16(acc1);
+    uint32x4_t s2 = vpaddlq_u16(acc2);
+    uint32x4_t s3 = vpaddlq_u16(acc3);
+    uint32x4_t s01 = vaddq_u32(s0, s1);
+    uint32x4_t s23 = vaddq_u32(s2, s3);
+    uint32x4_t st = vaddq_u32(s01, s23);
+    uint64x2_t s64 = vpaddlq_u32(st);
+    bits += (long long)(vgetq_lane_u64(s64, 0) + vgetq_lane_u64(s64, 1));
+
+    /* Remaining 64B blocks (keep vector domain) */
+    while (count >= 64) {
+        uint8x16_t d0 = vld1q_u8(p +  0);
+        uint8x16_t d1 = vld1q_u8(p + 16);
+        uint8x16_t d2 = vld1q_u8(p + 32);
+        uint8x16_t d3 = vld1q_u8(p + 48);
+
+        uint8x16_t c0 = vcntq_u8(d0);
+        uint8x16_t c1 = vcntq_u8(d1);
+        uint8x16_t c2 = vcntq_u8(d2);
+        uint8x16_t c3 = vcntq_u8(d3);
+
+        uint64x2_t t0 = vpaddlq_u32(vpaddlq_u16(vpaddlq_u8(c0)));
+        uint64x2_t t1 = vpaddlq_u32(vpaddlq_u16(vpaddlq_u8(c1)));
+        uint64x2_t t2 = vpaddlq_u32(vpaddlq_u16(vpaddlq_u8(c2)));
+        uint64x2_t t3 = vpaddlq_u32(vpaddlq_u16(vpaddlq_u8(c3)));
+
+        uint64x2_t s = vaddq_u64(vaddq_u64(t0, t1), vaddq_u64(t2, t3));
+        bits += (long long)(vgetq_lane_u64(s, 0) + vgetq_lane_u64(s, 1));
+
+        p += 64;
+        count -= 64;
+    }
+
+    /* 16B chunks */
+    while (count >= 16) {
+        uint8x16_t d = vld1q_u8(p);
+        uint64x2_t s = vpaddlq_u32(vpaddlq_u16(vpaddlq_u8(vcntq_u8(d))));
+        bits += (long long)(vgetq_lane_u64(s, 0) + vgetq_lane_u64(s, 1));
+        p += 16;
+        count -= 16;
+    }
+
+    /* Tail */
+    while (count--) bits += bitsinbyte[*p++];
+
+    return bits;
+}
+#endif
+
+#ifdef HAVE_AVX512
+/* AVX512 optimized version of redisPopcount using VPOPCNTDQ instruction.
+ * This function requires AVX512F and AVX512VPOPCNTDQ support. */
+ATTRIBUTE_TARGET_AVX512_POPCOUNT
+long long redisPopCountAvx512(void *s, long count) {
+    long long bits = 0;
+    unsigned char *p = s;
+
+    /* Align to 64-byte boundary for optimal AVX512 performance */
+    while ((unsigned long)p & 63 && count) {
+        bits += bitsinbyte[*p++];
+        count--;
+    }
+
+    /* Process 64 bytes at a time using AVX512 */
+    while (count >= 64) {
+        __m512i data = _mm512_loadu_si512((__m512i*)p);
+        __m512i popcnt = _mm512_popcnt_epi64(data);
+
+        /* Sum all 8 64-bit popcount results */
+        bits += _mm512_reduce_add_epi64(popcnt);
+
+        p += 64;
+        count -= 64;
+
+        /* Prefetch next cache line */
+        redis_prefetch_read(p + 2048);
+    }
+
+    /* Handle remaining bytes with scalar popcount */
+    while (count >= 8) {
+        bits += __builtin_popcountll(*(uint64_t*)p);
+        p += 8;
+        count -= 8;
+    }
+
+    /* Handle final bytes */
+    while (count--) {
+        bits += bitsinbyte[*p++];
+    }
+
+    return bits;
+}
+#endif
+
+#ifdef HAVE_AVX2
+/* AVX2 optimized version of redisPopcount.
+ * This function requires AVX2 and POPCNT support. */
+ATTRIBUTE_TARGET_AVX2_POPCOUNT
+long long redisPopCountAvx2(void *s, long count) {
+    long long bits = 0;
+    unsigned char *p = s;
+
+    /* Align to 8-byte boundary for 64-bit operations */
+    while ((unsigned long)p & 7 && count) {
+        bits += bitsinbyte[*p++];
+        count--;
+    }
+
+    /* Use separate counters to avoid dependencies, similar to regular redisPopcount */
+    uint64_t cnt[4];
+    memset(cnt, 0, sizeof(cnt));
+
+    /* Process 32 bytes at a time using POPCNT on 64-bit chunks */
+    while (count >= 32) {
+        cnt[0] += __builtin_popcountll(*(uint64_t*)(p));
+        cnt[1] += __builtin_popcountll(*(uint64_t*)(p + 8));
+        cnt[2] += __builtin_popcountll(*(uint64_t*)(p + 16));
+        cnt[3] += __builtin_popcountll(*(uint64_t*)(p + 24));
+
+        p += 32;
+        count -= 32;
+
+        /* Prefetch next cache line */
+        redis_prefetch_read(p + 2048);
+    }
+
+    bits += cnt[0] + cnt[1] + cnt[2] + cnt[3];
+
+    /* Handle remaining bytes with scalar popcount */
+    while (count >= 8) {
+        bits += __builtin_popcountll(*(uint64_t*)p);
+        p += 8;
+        count -= 8;
+    }
+
+    /* Handle final bytes */
+    while (count--) {
+        bits += bitsinbyte[*p++];
+    }
+
+    return bits;
+}
+#endif
+
+/* Automatically select the best available popcount implementation */
+static inline long long redisPopcountAuto(const unsigned char *p, long count) {
+#ifdef HAVE_AVX512
+    if (BITOP_USE_AVX512) {
+        return redisPopCountAvx512((void*)p, count);
+    }
+#endif
+#ifdef HAVE_AVX2
+    if (BITOP_USE_AVX2) {
+        return redisPopCountAvx2((void*)p, count);
+    }
+#endif
+#ifdef HAVE_AARCH64_NEON
+    return redisPopCountAarch64((void*)p, count);
+#else
+    return redisPopcount((void*)p, count);
+#endif
+}
+
 /* Return the position of the first bit set to one (if 'bit' is 1) or
  * zero (if 'bit' is 0) in the bitmap starting at 's' and long 'count' bytes.
  *
@@ -525,6 +784,7 @@ static kvobj *lookupStringForBitCommand(client *c, uint64_t maxbit,
 {
     dictEntryLink link;
     size_t byte = maxbit >> 3;
+    size_t oldAllocSize = 0;
     kvobj *o = lookupKeyWriteWithLink(c->db,c->argv[1],&link);
     if (checkType(c,o,OBJ_STRING)) return NULL;
 
@@ -536,7 +796,11 @@ static kvobj *lookupStringForBitCommand(client *c, uint64_t maxbit,
     } else {
         o = dbUnshareStringValue(c->db,c->argv[1],o);
         *strOldSize  = sdslen(o->ptr);
+        if (server.memory_tracking_per_slot)
+            oldAllocSize = stringObjectAllocSize(o);
         o->ptr = sdsgrowzero(o->ptr,byte+1);
+        if (server.memory_tracking_per_slot)
+            updateSlotAllocSize(c->db, getKeySlot(c->argv[1]->ptr), oldAllocSize, stringObjectAllocSize(o));
         *strGrowSize = sdslen(o->ptr) - *strOldSize;
     }
     return o;
@@ -659,7 +923,7 @@ void getbitCommand(client *c) {
  * 256-bit registers so if `minlen` is not a multiple of 32 some of the bytes
  * will be skipped. They will be taken care for in the unoptimized loop in the
  * main bitopCommand function. */
-ATTRIBUTE_TARGET_AVX2
+ATTRIBUTE_TARGET_AVX2_POPCOUNT
 unsigned long bitopCommandAVX(unsigned char **keys, unsigned char *res, 
                               unsigned long op, unsigned long numkeys,
                               unsigned long minlen)
@@ -1268,7 +1532,11 @@ void bitcountCommand(client *c) {
         addReply(c,shared.czero);
     } else {
         long bytes = (long)(end-start+1);
-        long long count = redisPopcount(p+start,bytes);
+        long long count;
+
+        /* Use the best available popcount implementation */
+        count = redisPopcountAuto(p+start, bytes);
+
         if (first_byte_neg_mask != 0 || last_byte_neg_mask != 0) {
             unsigned char firstlast[2] = {0, 0};
             /* We may count bits of first byte and last byte which are out of
@@ -1276,7 +1544,9 @@ void bitcountCommand(client *c) {
             * bits in the range to zero. So these bit will not be excluded. */
             if (first_byte_neg_mask != 0) firstlast[0] = p[start] & first_byte_neg_mask;
             if (last_byte_neg_mask != 0) firstlast[1] = p[end] & last_byte_neg_mask;
-            count -= redisPopcount(firstlast,2);
+
+            /* Use the same popcount implementation for consistency */
+            count -= redisPopcountAuto(firstlast, 2);
         }
         addReplyLongLong(c,count);
     }
@@ -1695,3 +1965,73 @@ void bitfieldCommand(client *c) {
 void bitfieldroCommand(client *c) {
     bitfieldGeneric(c, BITFIELD_FLAG_READONLY);
 }
+
+#ifdef REDIS_TEST
+/* Test function to verify popcount implementations */
+int bitopsTest(int argc, char **argv, int flags) {
+    UNUSED(argc);
+    UNUSED(argv);
+    UNUSED(flags);
+
+    /* Test data with known popcount values */
+    unsigned char test_data[] = {0xFF, 0x00, 0xAA, 0x55, 0xF0, 0x0F, 0x33, 0xCC};
+    int expected_bits = 8 + 0 + 4 + 4 + 4 + 4 + 4 + 4; /* = 32 bits */
+
+    long long result_regular = redisPopcount(test_data, sizeof(test_data));
+
+    printf("Regular popcount: %lld (expected: %d)\n", result_regular, expected_bits);
+
+    if (result_regular != expected_bits) {
+        printf("FAIL: Regular popcount mismatch\n");
+        return 1;
+    }
+
+#ifdef HAVE_AVX2
+    if (BITOP_USE_AVX2) {
+        long long result_avx2 = redisPopCountAvx2(test_data, sizeof(test_data));
+        printf("AVX2 popcount: %lld (expected: %d)\n", result_avx2, expected_bits);
+
+        if (result_avx2 != expected_bits) {
+            printf("FAIL: AVX2 popcount mismatch\n");
+            return 1;
+        }
+    } else {
+        printf("AVX2 not supported on this CPU\n");
+    }
+#else
+    printf("AVX2 not compiled in\n");
+#endif
+
+#ifdef HAVE_AVX512
+    if (BITOP_USE_AVX512) {
+        long long result_avx512 = redisPopCountAvx512(test_data, sizeof(test_data));
+        printf("AVX512 popcount: %lld (expected: %d)\n", result_avx512, expected_bits);
+
+        if (result_avx512 != expected_bits) {
+            printf("FAIL: AVX512 popcount mismatch\n");
+            return 1;
+        }
+    } else {
+        printf("AVX512 not supported on this CPU\n");
+    }
+#else
+    printf("AVX512 not compiled in\n");
+#endif
+
+#ifdef HAVE_AARCH64_NEON
+    {
+        long long result_aarch64 = redisPopCountAarch64(test_data, sizeof(test_data));
+        printf("AArch64 NEON popcount: %lld (expected: %d)\n", result_aarch64, expected_bits);
+
+        if (result_aarch64 != expected_bits) {
+            printf("FAIL: AArch64 NEON popcount mismatch\n");
+            return 1;
+        }
+    }
+#else
+    printf("AArch64 NEON not available\n");
+#endif
+    printf("All popcount tests passed!\n");
+    return 0;
+}
+#endif
