@@ -14,6 +14,8 @@
 #include "bio.h"
 #include "atomicvar.h"
 #include "script.h"
+#include "cluster.h"
+#include "cluster_asm.h"
 #include <math.h>
 
 /* ----------------------------------------------------------------------------
@@ -80,6 +82,12 @@ unsigned long long estimateObjectIdleTime(robj *o) {
     }
 }
 
+/* During atomic slot migration, keys that are being imported are in an
+ * intermediate state. We cannot evict them and therefore skip them. */
+static int randomEvictionShouldSkipDictIndex(int didx) {
+    return !clusterCanAccessKeysInSlot(didx);
+}
+
 /* LRU approximation algorithm
  *
  * Redis uses an approximation of the LRU algorithm that runs in constant
@@ -127,7 +135,9 @@ int evictionPoolPopulate(redisDb *db, kvstore *samplekvs, struct evictionPoolEnt
     int j, k, count;
     dictEntry *samples[server.maxmemory_samples];
 
-    int slot = kvstoreGetFairRandomDictIndex(samplekvs);
+    /* Don't retry, since we will call evictionPoolPopulate multiple times if needed. */
+    int slot = kvstoreGetFairRandomDictIndex(samplekvs, randomEvictionShouldSkipDictIndex, 1, 0);
+    if (slot == -1) return 0;
     count = kvstoreDictGetSomeKeys(samplekvs,slot,samples,server.maxmemory_samples);
     for (j = 0; j < count; j++) {
         unsigned long long idle;
@@ -336,6 +346,11 @@ size_t freeMemoryGetNotCountedMemory(void) {
         }
     }
 
+    /* The migrate client is like a replica, we also push DELs into it when
+     * evicting keys belonging to the migrating slot, so we don't count its
+     * output buffer to avoid eviction loop. */
+    overhead += asmGetMigrateOutputBufferSize();
+
     if (server.aof_state != AOF_OFF) {
         overhead += sdsAllocSize(server.aof_buf);
     }
@@ -459,6 +474,13 @@ static int isSafeToPerformEvictions(void) {
      * and just be masters exact copies. */
     if (server.masterhost && server.repl_slave_ignore_maxmemory) return 0;
 
+    /* Disable eviction during slot migration import to avoid delays and errors
+     * caused by failed evictions. A special client is created for data import,
+     * identified by the CLIENT_MASTER and CLIENT_ASM_IMPORTING flags. */
+    if (server.current_client && server.current_client->flags & CLIENT_MASTER &&
+        server.current_client->flags & CLIENT_ASM_IMPORTING)
+        return 0;
+
     /* If 'evict' action is paused, for whatever reason, then return false */
     if (isPausedActionsWithUpdate(PAUSE_ACTION_EVICT)) return 0;
 
@@ -556,6 +578,7 @@ int performEvictions(void) {
             struct evictionPoolEntry *pool = EvictionPoolLRU;
             while (bestkey == NULL) {
                 unsigned long total_keys = 0;
+                unsigned long total_sampled_keys = 0;
 
                 /* We don't want to make local-db choices when expiring keys,
                  * so to start populate the eviction pool sampling keys from
@@ -577,6 +600,7 @@ int performEvictions(void) {
                     /* Do not exceed the number of non-empty slots when looping. */
                     while (l--) {
                         sampled_keys += evictionPoolPopulate(db, kvs, pool);
+                        total_sampled_keys += sampled_keys;
                         /* We have sampled enough keys in the current db, exit the loop. */
                         if (sampled_keys >= (unsigned long) server.maxmemory_samples)
                             break;
@@ -589,6 +613,10 @@ int performEvictions(void) {
                 }
                 if (!total_keys) break; /* No keys to evict. */
 
+                /* If we iterated all the DBs and all non-empty slot dicts, then
+                 * did not sample any key, stop sampling. */
+                if (!total_sampled_keys) break;
+
                 /* Go backward from best to worst element to evict. */
                 for (k = EVPOOL_SIZE-1; k >= 0; k--) {
                     if (pool[k].key == NULL) continue;
@@ -636,7 +664,8 @@ int performEvictions(void) {
                 } else {
                     kvs = db->expires;
                 }
-                int slot = kvstoreGetFairRandomDictIndex(kvs);
+                int slot = kvstoreGetFairRandomDictIndex(kvs, randomEvictionShouldSkipDictIndex, 16, 0);
+                if (slot == -1) continue;
                 de = kvstoreDictGetRandomKey(kvs, slot);
                 if (de) {
                     kvobj *kv = dictGetKV(de);
