@@ -23,7 +23,7 @@
  * to ensure the safe invocation of 'zmalloc_usable_size().
  * See comment in zmalloc_usable_size(). */
 #define lp_malloc(sz) zmalloc_usable(sz,NULL)
-#define lp_realloc(ptr,sz) zrealloc_usable(ptr,sz,NULL)
+#define lp_realloc(ptr,sz) zrealloc_usable(ptr,sz,NULL,NULL)
 #define lp_free zfree
 #define lp_malloc_size zmalloc_usable_size
 #endif
