@@ -482,7 +482,7 @@ int pubsubPublishMessageInternal(robj *channel, robj *message, pubsubtype type)
         while ((entry = dictNext(&iter)) != NULL) {
             client *c = dictGetKey(entry);
             addReplyPubsubMessage(c,channel,message,*type.messageBulk);
-            if (server.cluster_enabled && server.cluster_slot_stats_enabled)
+            if (clusterSlotStatsEnabled())
                 clusterSlotStatsAddNetworkBytesOutForShardedPubSubInternalPropagation(c, slot);
             updateClientMemUsageAndBucket(c);
             receivers++;
