@@ -132,6 +132,7 @@ int setTypeAddAux(robj *set, char *str, size_t len, int64_t llval, int str_is_sd
             /* Key doesn't already exist in the set. Add it but dup the key. */
             if (sdsval == str) sdsval = sdsdup(sdsval);
             dictSetKeyAtLink(ht, sdsval, &bucket, 1);
+            *htGetMetadataSize(ht) += sdsAllocSize(sdsval);
             return 1;
         } else if (sdsval != str) {
             /* String is already a member. Free our temporary sds copy. */
@@ -160,7 +161,9 @@ int setTypeAddAux(robj *set, char *str, size_t len, int64_t llval, int str_is_sd
             } else {
                 /* Size limit is reached. Convert to hashtable and add. */
                 setTypeConvertAndExpand(set, OBJ_ENCODING_HT, lpLength(lp) + 1, 1);
-                serverAssert(dictAdd(set->ptr,sdsnewlen(str,len),NULL) == DICT_OK);
+                sds newval = sdsnewlen(str,len);
+                serverAssert(dictAdd(set->ptr,newval,NULL) == DICT_OK);
+                *htGetMetadataSize(set->ptr) += sdsAllocSize(newval);
             }
             return 1;
         }
@@ -204,7 +207,9 @@ int setTypeAddAux(robj *set, char *str, size_t len, int64_t llval, int str_is_sd
                                         intsetLen(set->ptr) + 1, 1);
                 /* The set *was* an intset and this value is not integer
                  * encodable, so dictAdd should always work. */
-                serverAssert(dictAdd(set->ptr,sdsnewlen(str,len),NULL) == DICT_OK);
+                sds newval = sdsnewlen(str,len);
+                serverAssert(dictAdd(set->ptr,newval,NULL) == DICT_OK);
+                *htGetMetadataSize(set->ptr) += sdsAllocSize(newval);
                 return 1;
             }
         }
@@ -474,6 +479,22 @@ unsigned long setTypeSize(const robj *subject) {
     }
 }
 
+size_t setTypeAllocSize(const robj *o) {
+    serverAssertWithInfo(NULL,o,o->type == OBJ_SET);
+    size_t size = 0;
+    if (o->encoding == OBJ_ENCODING_HT) {
+        dict *d = o->ptr;
+        size += sizeof(dict) + dictMemUsage(d) + *htGetMetadataSize(d);
+    } else if (o->encoding == OBJ_ENCODING_INTSET) {
+        size = intsetAllocSize(o->ptr);
+    } else if (o->encoding == OBJ_ENCODING_LISTPACK) {
+        size = lpBytes(o->ptr);
+    } else {
+        serverPanic("Unknown set encoding");
+    }
+    return size;
+}
+
 /* Convert the set to specified encoding. The resulting dict (when converting
  * to a hash table) is presized to hold the number of elements in the original
  * set. */
@@ -503,9 +524,11 @@ int setTypeConvertAndExpand(robj *setobj, int enc, unsigned long cap, int panic)
         }
 
         /* To add the elements we extract integers and create redis objects */
+        size_t *alloc_size = htGetMetadataSize(d);
         si = setTypeInitIterator(setobj);
         while ((element = setTypeNextObject(si)) != NULL) {
             serverAssert(dictAdd(d,element,NULL) == DICT_OK);
+            *alloc_size += sdsAllocSize(element);
         }
         setTypeReleaseIterator(si);
 
@@ -591,6 +614,7 @@ void saddCommand(client *c) {
     kvobj *set;
     int j, added = 0;
     dictEntryLink link;
+    size_t oldsize = 0;
 
     set = lookupKeyWriteWithLink(c->db,c->argv[1], &link);
     if (checkType(c,set,OBJ_SET)) return;
@@ -599,12 +623,20 @@ void saddCommand(client *c) {
         robj *o = setTypeCreate(c->argv[2]->ptr, c->argc - 2);
         set = dbAddByLink(c->db, c->argv[1], &o, &link);
     } else {
+        if (server.memory_tracking_per_slot)
+            oldsize = setTypeAllocSize(set);
         setTypeMaybeConvert(set, c->argc - 2);
+        if (server.memory_tracking_per_slot)
+            updateSlotAllocSize(c->db, getKeySlot(c->argv[1]->ptr), oldsize, setTypeAllocSize(set));
     }
 
+    if (server.memory_tracking_per_slot)
+        oldsize = setTypeAllocSize(set);
     for (j = 2; j < c->argc; j++) {
         if (setTypeAdd(set,c->argv[j]->ptr)) added++;
     }
+    if (server.memory_tracking_per_slot)
+        updateSlotAllocSize(c->db, getKeySlot(c->argv[1]->ptr), oldsize, setTypeAllocSize(set));
     if (added) {
         unsigned long size = setTypeSize(set);
         updateKeysizesHist(c->db, getKeySlot(c->argv[1]->ptr), OBJ_SET, size - added, size);
@@ -617,23 +649,30 @@ void saddCommand(client *c) {
 
 void sremCommand(client *c) {
     int j, deleted = 0, keyremoved = 0;
+    size_t oldsize = 0;
 
     kvobj *set = lookupKeyWriteOrReply(c, c->argv[1], shared.czero);
     if (set == NULL || checkType(c, set, OBJ_SET))
         return;
 
     unsigned long oldSize = setTypeSize(set);
+    if (server.memory_tracking_per_slot)
+        oldsize = setTypeAllocSize(set);
 
     for (j = 2; j < c->argc; j++) {
         if (setTypeRemove(set,c->argv[j]->ptr)) {
             deleted++;
             if (setTypeSize(set) == 0) {
+                if (server.memory_tracking_per_slot)
+                    updateSlotAllocSize(c->db, getKeySlot(c->argv[1]->ptr), oldsize, setTypeAllocSize(set));
                 dbDeleteSkipKeysizesUpdate(c->db, c->argv[1]);
                 keyremoved = 1;
                 break;
             }
         }
     }
+    if (server.memory_tracking_per_slot && !keyremoved)
+        updateSlotAllocSize(c->db, getKeySlot(c->argv[1]->ptr), oldsize, setTypeAllocSize(set));
     if (deleted) {
         int64_t newSize = oldSize - deleted;
 
@@ -652,6 +691,7 @@ void sremCommand(client *c) {
 
 void smoveCommand(client *c) {
     robj *srcset, *dstset, *ele;
+    size_t oldSrcAllocSize = 0, oldDstAllocSize = 0;
     srcset = lookupKeyWrite(c->db,c->argv[1]);
     dstset = lookupKeyWrite(c->db,c->argv[2]);
     ele = c->argv[3];
@@ -674,8 +714,13 @@ void smoveCommand(client *c) {
         return;
     }
 
+    if (server.memory_tracking_per_slot)
+        oldSrcAllocSize = setTypeAllocSize(srcset);
+    int deleted = setTypeRemove(srcset,ele->ptr);
+    if (server.memory_tracking_per_slot)
+        updateSlotAllocSize(c->db, getKeySlot(c->argv[1]->ptr), oldSrcAllocSize, setTypeAllocSize(srcset));
     /* If the element cannot be removed from the src set, return 0. */
-    if (!setTypeRemove(srcset,ele->ptr)) {
+    if (!deleted) {
         addReply(c,shared.czero);
         return;
     }
@@ -701,6 +746,8 @@ void smoveCommand(client *c) {
     signalModifiedKey(c,c->db,c->argv[1]);
     server.dirty++;
 
+    if (server.memory_tracking_per_slot)
+        oldDstAllocSize = setTypeAllocSize(dstset);
     /* An extra key has changed when ele was successfully added to dstset */
     if (setTypeAdd(dstset,ele->ptr)) {
         unsigned long dstLen = setTypeSize(dstset);
@@ -709,35 +756,47 @@ void smoveCommand(client *c) {
         signalModifiedKey(c,c->db,c->argv[2]);
         notifyKeyspaceEvent(NOTIFY_SET,"sadd",c->argv[2],c->db->id);
     }
+    if (server.memory_tracking_per_slot)
+        updateSlotAllocSize(c->db, getKeySlot(c->argv[2]->ptr), oldDstAllocSize, setTypeAllocSize(dstset));
     addReply(c,shared.cone);
 }
 
 void sismemberCommand(client *c) {
     kvobj *set;
+    size_t oldsize = 0;
 
     if ((set = lookupKeyReadOrReply(c,c->argv[1],shared.czero)) == NULL ||
         checkType(c,set,OBJ_SET)) return;
 
+    if (server.memory_tracking_per_slot)
+        oldsize = setTypeAllocSize(set);
     if (setTypeIsMember(set,c->argv[2]->ptr))
         addReply(c,shared.cone);
     else
         addReply(c,shared.czero);
+    if (server.memory_tracking_per_slot)
+        updateSlotAllocSize(c->db, getKeySlot(c->argv[1]->ptr), oldsize, setTypeAllocSize(set));
 }
 
 void smismemberCommand(client *c) {
     /* Don't abort when the key cannot be found. Non-existing keys are empty
      * sets, where SMISMEMBER should respond with a series of zeros. */
+    size_t oldsize = 0;
     kvobj *set = lookupKeyRead(c->db, c->argv[1]);
     if (set && checkType(c,set,OBJ_SET)) return;
 
     addReplyArrayLen(c,c->argc - 2);
 
+    if (server.memory_tracking_per_slot && set)
+        setTypeAllocSize(set);
     for (int j = 2; j < c->argc; j++) {
         if (set && setTypeIsMember(set,c->argv[j]->ptr))
             addReply(c,shared.cone);
         else
             addReply(c,shared.czero);
     }
+    if (server.memory_tracking_per_slot && set)
+        updateSlotAllocSize(c->db, getKeySlot(c->argv[1]->ptr), oldsize, setTypeAllocSize(set));
 }
 
 void scardCommand(client *c) {
@@ -760,6 +819,7 @@ void scardCommand(client *c) {
 void spopWithCountCommand(client *c) {
     long l;
     unsigned long count, size, toRemove;
+    size_t oldsize = 0;
 
     /* Get the count argument */
     if (getPositiveLongFromObjectOrReply(c,c->argv[2],&l,NULL) != C_OK) return;
@@ -831,6 +891,8 @@ void spopWithCountCommand(client *c) {
         set->encoding == OBJ_ENCODING_LISTPACK)
     {
         /* Specialized case for listpack. Traverse it only once. */
+        if (server.memory_tracking_per_slot)
+            oldsize = setTypeAllocSize(set);
         unsigned char *lp = set->ptr;
         unsigned char *p = lpFirst(lp);
         unsigned int index = 0;
@@ -865,7 +927,11 @@ void spopWithCountCommand(client *c) {
         zfree(ps);
         set->ptr = lp;
         updateKeysizesHist(c->db, getKeySlot(c->argv[1]->ptr), OBJ_SET, size, size - count);
+        if (server.memory_tracking_per_slot)
+            updateSlotAllocSize(c->db, getKeySlot(c->argv[1]->ptr), oldsize, setTypeAllocSize(set));
     } else if (remaining*SPOP_MOVE_STRATEGY_MUL > count) {
+        if (server.memory_tracking_per_slot)
+            oldsize = setTypeAllocSize(set);
         for (unsigned long i = 0; i < count; i++) {
             propargv[propindex] = setTypePopRandom(set);
             addReplyBulk(c, propargv[propindex]);
@@ -880,6 +946,8 @@ void spopWithCountCommand(client *c) {
             }
         }
         updateKeysizesHist(c->db, getKeySlot(c->argv[1]->ptr), OBJ_SET, size, size - count);
+        if (server.memory_tracking_per_slot)
+            updateSlotAllocSize(c->db, getKeySlot(c->argv[1]->ptr), oldsize, setTypeAllocSize(set));
     } else {
     /* CASE 3: The number of elements to return is very big, approaching
      * the size of the set itself. After some time extracting random elements
@@ -890,6 +958,8 @@ void spopWithCountCommand(client *c) {
      * set). Then we return the elements left in the original set and
      * release it. */
         robj *newset = NULL;
+        if (server.memory_tracking_per_slot)
+            oldsize = setTypeAllocSize(set);
 
         /* Create a new set with just the remaining elements. */
         if (set->encoding == OBJ_ENCODING_LISTPACK) {
@@ -949,6 +1019,8 @@ void spopWithCountCommand(client *c) {
          * but here we're building the new set from the existing one. As a result, 
          * the size of the old set has already changed by the time we reach this point. */
         updateKeysizesHist(c->db, getKeySlot(c->argv[1]->ptr), OBJ_SET, size, size-count);
+        if (server.memory_tracking_per_slot)
+            updateSlotAllocSize(c->db, getKeySlot(c->argv[1]->ptr), oldsize, setTypeAllocSize(set));
         dbReplaceValue(c->db, c->argv[1], &newset, 0);
     }
 
@@ -973,6 +1045,7 @@ void spopWithCountCommand(client *c) {
 void spopCommand(client *c) {
     unsigned long size;
     robj *ele;
+    size_t oldsize = 0;
 
     if (c->argc == 3) {
         spopWithCountCommand(c);
@@ -990,9 +1063,15 @@ void spopCommand(client *c) {
     size = setTypeSize(kv);
     updateKeysizesHist(c->db, getKeySlot(c->argv[1]->ptr), OBJ_SET, size, size-1);
 
+    if (server.memory_tracking_per_slot)
+        oldsize = setTypeAllocSize(kv);
+
     /* Pop a random element from the kv */
     ele = setTypePopRandom(kv);
 
+    if (server.memory_tracking_per_slot)
+        updateSlotAllocSize(c->db, getKeySlot(c->argv[1]->ptr), oldsize, setTypeAllocSize(kv));
+
     notifyKeyspaceEvent(NOTIFY_SET,"spop",c->argv[1],c->db->id);
 
     /* Replicate/AOF this command as an SREM operation */
@@ -1236,6 +1315,7 @@ void srandmemberCommand(client *c) {
     char *str;
     size_t len = 0;
     int64_t llele = 0;
+    size_t oldsize = 0;
 
     if (c->argc == 3) {
         srandmemberWithCountCommand(c);
@@ -1249,7 +1329,11 @@ void srandmemberCommand(client *c) {
     if ((set = lookupKeyReadOrReply(c,c->argv[1],shared.null[c->resp]))
         == NULL || checkType(c,set,OBJ_SET)) return;
 
+    if (server.memory_tracking_per_slot)
+        oldsize = setTypeAllocSize(set);
     setTypeRandomElement(set, &str, &len, &llele);
+    if (server.memory_tracking_per_slot)
+        updateSlotAllocSize(c->db, getKeySlot(c->argv[1]->ptr), oldsize, setTypeAllocSize(set));
     if (str == NULL) {
         addReplyBulkLongLong(c,llele);
     } else {
@@ -1257,16 +1341,22 @@ void srandmemberCommand(client *c) {
     }
 }
 
+typedef struct setopsrc {
+    robj *set;
+    size_t oldsize;
+} setopsrc;
+
 int qsortCompareSetsByCardinality(const void *s1, const void *s2) {
-    if (setTypeSize(*(robj**)s1) > setTypeSize(*(robj**)s2)) return 1;
-    if (setTypeSize(*(robj**)s1) < setTypeSize(*(robj**)s2)) return -1;
+    robj *o1 = ((setopsrc*)s1)->set, *o2 = ((setopsrc*)s2)->set;
+    if (setTypeSize(o1) > setTypeSize(o2)) return 1;
+    if (setTypeSize(o1) < setTypeSize(o2)) return -1;
     return 0;
 }
 
 /* This is used by SDIFF and in this case we can receive NULL that should
  * be handled as empty sets. */
 int qsortCompareSetsByRevCardinality(const void *s1, const void *s2) {
-    robj *o1 = *(robj**)s1, *o2 = *(robj**)s2;
+    robj *o1 = ((setopsrc*)s1)->set, *o2 = ((setopsrc*)s2)->set;
     unsigned long first = o1 ? setTypeSize(o1) : 0;
     unsigned long second = o2 ? setTypeSize(o2) : 0;
 
@@ -1286,7 +1376,7 @@ int qsortCompareSetsByRevCardinality(const void *s1, const void *s2) {
 void sinterGenericCommand(client *c, robj **setkeys,
                           unsigned long setnum, robj *dstkey,
                           int cardinality_only, unsigned long limit) {
-    kvobj **sets = zmalloc(sizeof(robj*)*setnum);
+    setopsrc *sets = zmalloc(sizeof(setopsrc)*setnum);
     setTypeIterator *si;
     robj *dstset = NULL;
     char *str;
@@ -1301,14 +1391,17 @@ void sinterGenericCommand(client *c, robj **setkeys,
         if (!kv) {
             /* A NULL is considered an empty set */
             empty += 1;
-            sets[j] = NULL;
+            sets[j].set = NULL;
+            sets[j].oldsize = 0;
             continue;
         }
         if (checkType(c, kv, OBJ_SET)) {
             zfree(sets);
             return;
         }
-        sets[j] = kv;
+        sets[j].set = kv;
+        if (server.memory_tracking_per_slot)
+            sets[j].oldsize = setTypeAllocSize(kv);
     }
 
     /* Set intersection with an empty set always results in an empty set.
@@ -1332,7 +1425,7 @@ void sinterGenericCommand(client *c, robj **setkeys,
 
     /* Sort sets from the smallest to largest, this will improve our
      * algorithm's performance */
-    qsort(sets,setnum,sizeof(robj*),qsortCompareSetsByCardinality);
+    qsort(sets,setnum,sizeof(setopsrc),qsortCompareSetsByCardinality);
 
     /* The first thing we should output is the total number of elements...
      * since this is a multi-bulk write, but at this stage we don't know
@@ -1342,16 +1435,16 @@ void sinterGenericCommand(client *c, robj **setkeys,
     if (dstkey) {
         /* If we have a target key where to store the resulting set
          * create this key with an empty set inside */
-        if (sets[0]->encoding == OBJ_ENCODING_INTSET) {
+        if (sets[0].set->encoding == OBJ_ENCODING_INTSET) {
             /* The first set is an intset, so the result is an intset too. The
              * elements are inserted in ascending order which is efficient in an
              * intset. */
             dstset = createIntsetObject();
-        } else if (sets[0]->encoding == OBJ_ENCODING_LISTPACK) {
+        } else if (sets[0].set->encoding == OBJ_ENCODING_LISTPACK) {
             /* To avoid many reallocs, we estimate that the result is a listpack
              * of approximately the same size as the first set. Then we shrink
              * it or possibly convert it to intset in the end. */
-            unsigned char *lp = lpNew(lpBytes(sets[0]->ptr));
+            unsigned char *lp = lpNew(lpBytes(sets[0].set->ptr));
             dstset = createObject(OBJ_SET, lp);
             dstset->encoding = OBJ_ENCODING_LISTPACK;
         } else {
@@ -1368,11 +1461,11 @@ void sinterGenericCommand(client *c, robj **setkeys,
      * the element against all the other sets, if at least one set does
      * not include the element it is discarded */
     int only_integers = 1;
-    si = setTypeInitIterator(sets[0]);
+    si = setTypeInitIterator(sets[0].set);
     while((encoding = setTypeNext(si, &str, &len, &intobj)) != -1) {
         for (j = 1; j < setnum; j++) {
-            if (sets[j] == sets[0]) continue;
-            if (!setTypeIsMemberAux(sets[j], str, len, intobj,
+            if (sets[j].set == sets[0].set) continue;
+            if (!setTypeIsMemberAux(sets[j].set, str, len, intobj,
                                     encoding == OBJ_ENCODING_HT))
                 break;
         }
@@ -1414,6 +1507,14 @@ void sinterGenericCommand(client *c, robj **setkeys,
     }
     setTypeReleaseIterator(si);
 
+    if (server.memory_tracking_per_slot) {
+        for (j = 0; j < setnum; j++) {
+            if (!sets[j].set) continue;
+            updateSlotAllocSize(c->db, getKeySlot(setkeys[j]->ptr),
+                            sets[j].oldsize, setTypeAllocSize(sets[j].set));
+        }
+    }
+
     if (cardinality_only) {
         addReplyLongLong(c,cardinality);
     } else if (dstkey) {
@@ -1457,6 +1558,7 @@ void smembersCommand(client *c) {
     char *str;
     size_t len = 0;
     int64_t intobj = 0;
+    size_t oldsize = 0;
     kvobj *setobj = lookupKeyRead(c->db, c->argv[1]);
     if (checkType(c,setobj,OBJ_SET)) return;
     if (!setobj) {
@@ -1467,6 +1569,8 @@ void smembersCommand(client *c) {
     /* Prepare the response. */
     unsigned long length = setTypeSize(setobj);
     addReplySetLen(c,length);
+    if (server.memory_tracking_per_slot)
+        oldsize = setTypeAllocSize(setobj);
     /* Iterate through the elements of the set. */
     si = setTypeInitIterator(setobj);
 
@@ -1478,6 +1582,8 @@ void smembersCommand(client *c) {
         length--;
     }
     setTypeReleaseIterator(si);
+    if (server.memory_tracking_per_slot)
+        updateSlotAllocSize(c->db, getKeySlot(c->argv[1]->ptr), oldsize, setTypeAllocSize(setobj));
     serverAssert(length == 0); /* fail on corrupt data */
 }
 
@@ -1520,7 +1626,7 @@ void sinterstoreCommand(client *c) {
 
 void sunionDiffGenericCommand(client *c, robj **setkeys, int setnum,
                               robj *dstkey, int op) {
-    robj **sets = zmalloc(sizeof(robj*)*setnum);
+    setopsrc *sets = zmalloc(sizeof(setopsrc)*setnum);
     setTypeIterator *si;
     robj *dstset = NULL;
     int dstset_encoding = OBJ_ENCODING_INTSET;
@@ -1535,7 +1641,8 @@ void sunionDiffGenericCommand(client *c, robj **setkeys, int setnum,
     for (j = 0; j < setnum; j++) {
         kvobj *setobj = lookupKeyRead(c->db, setkeys[j]);
         if (!setobj) {
-            sets[j] = NULL;
+            sets[j].set = NULL;
+            sets[j].oldsize = 0;
             continue;
         }
         if (checkType(c,setobj,OBJ_SET)) {
@@ -1559,8 +1666,10 @@ void sunionDiffGenericCommand(client *c, robj **setkeys, int setnum,
             (setobj->encoding == OBJ_ENCODING_LISTPACK || setobj->encoding == OBJ_ENCODING_HT)) {
             dstset_encoding = OBJ_ENCODING_HT;
         }
-        sets[j] = setobj;
-        if (j > 0 && sets[0] == sets[j]) {
+        sets[j].set = setobj;
+        if (server.memory_tracking_per_slot)
+            sets[j].oldsize = setTypeAllocSize(setobj);
+        if (j > 0 && sets[0].set == sets[j].set) {
             sameset = 1; 
         }
     }
@@ -1574,14 +1683,14 @@ void sunionDiffGenericCommand(client *c, robj **setkeys, int setnum,
      * the sets.
      *
      * We compute what is the best bet with the current input here. */
-    if (op == SET_OP_DIFF && sets[0] && !sameset) {
+    if (op == SET_OP_DIFF && sets[0].set && !sameset) {
         long long algo_one_work = 0, algo_two_work = 0;
 
         for (j = 0; j < setnum; j++) {
-            if (sets[j] == NULL) continue;
+            if (sets[j].set == NULL) continue;
 
-            algo_one_work += setTypeSize(sets[0]);
-            algo_two_work += setTypeSize(sets[j]);
+            algo_one_work += setTypeSize(sets[0].set);
+            algo_two_work += setTypeSize(sets[j].set);
         }
 
         /* Algorithm 1 has better constant times and performs less operations
@@ -1593,7 +1702,7 @@ void sunionDiffGenericCommand(client *c, robj **setkeys, int setnum,
             /* With algorithm 1 it is better to order the sets to subtract
              * by decreasing size, so that we are more likely to find
              * duplicated elements ASAP. */
-            qsort(sets+1,setnum-1,sizeof(robj*),
+            qsort(sets+1,setnum-1,sizeof(setopsrc),
                 qsortCompareSetsByRevCardinality);
         }
     }
@@ -1611,9 +1720,9 @@ void sunionDiffGenericCommand(client *c, robj **setkeys, int setnum,
         /* Union is trivial, just add every element of every set to the
          * temporary set. */
         for (j = 0; j < setnum; j++) {
-            if (!sets[j]) continue; /* non existing keys are like empty sets */
+            if (!sets[j].set) continue; /* non existing keys are like empty sets */
 
-            si = setTypeInitIterator(sets[j]);
+            si = setTypeInitIterator(sets[j].set);
             while ((encoding = setTypeNext(si, &str, &len, &llval)) != -1) {
                 cardinality += setTypeAddAux(dstset, str, len, llval, encoding == OBJ_ENCODING_HT);
             }
@@ -1621,7 +1730,7 @@ void sunionDiffGenericCommand(client *c, robj **setkeys, int setnum,
         }
     } else if (op == SET_OP_DIFF && sameset) {
         /* At least one of the sets is the same one (same key) as the first one, result must be empty. */
-    } else if (op == SET_OP_DIFF && sets[0] && diff_algo == 1) {
+    } else if (op == SET_OP_DIFF && sets[0].set && diff_algo == 1) {
         /* DIFF Algorithm 1:
          *
          * We perform the diff by iterating all the elements of the first set,
@@ -1630,12 +1739,12 @@ void sunionDiffGenericCommand(client *c, robj **setkeys, int setnum,
          *
          * This way we perform at max N*M operations, where N is the size of
          * the first set, and M the number of sets. */
-        si = setTypeInitIterator(sets[0]);
+        si = setTypeInitIterator(sets[0].set);
         while ((encoding = setTypeNext(si, &str, &len, &llval)) != -1) {
             for (j = 1; j < setnum; j++) {
-                if (!sets[j]) continue; /* no key is an empty set. */
-                if (sets[j] == sets[0]) break; /* same set! */
-                if (setTypeIsMemberAux(sets[j], str, len, llval,
+                if (!sets[j].set) continue; /* no key is an empty set. */
+                if (sets[j].set == sets[0].set) break; /* same set! */
+                if (setTypeIsMemberAux(sets[j].set, str, len, llval,
                                        encoding == OBJ_ENCODING_HT))
                     break;
             }
@@ -1645,7 +1754,7 @@ void sunionDiffGenericCommand(client *c, robj **setkeys, int setnum,
             }
         }
         setTypeReleaseIterator(si);
-    } else if (op == SET_OP_DIFF && sets[0] && diff_algo == 2) {
+    } else if (op == SET_OP_DIFF && sets[0].set && diff_algo == 2) {
         /* DIFF Algorithm 2:
          *
          * Add all the elements of the first set to the auxiliary set.
@@ -1654,9 +1763,9 @@ void sunionDiffGenericCommand(client *c, robj **setkeys, int setnum,
          * This is O(N) where N is the sum of all the elements in every
          * set. */
         for (j = 0; j < setnum; j++) {
-            if (!sets[j]) continue; /* non existing keys are like empty sets */
+            if (!sets[j].set) continue; /* non existing keys are like empty sets */
 
-            si = setTypeInitIterator(sets[j]);
+            si = setTypeInitIterator(sets[j].set);
             while((encoding = setTypeNext(si, &str, &len, &llval)) != -1) {
                 if (j == 0) {
                     cardinality += setTypeAddAux(dstset, str, len, llval,
@@ -1673,6 +1782,13 @@ void sunionDiffGenericCommand(client *c, robj **setkeys, int setnum,
             if (cardinality == 0) break;
         }
     }
+    if (server.memory_tracking_per_slot) {
+        for (j = 0; j < setnum; j++) {
+            if (!sets[j].set) continue;
+            updateSlotAllocSize(c->db, getKeySlot(setkeys[j]->ptr),
+                            sets[j].oldsize, setTypeAllocSize(sets[j].set));
+        }
+    }
 
     /* Output the content of the resulting set, if not in STORE mode */
     if (!dstkey) {
@@ -1733,9 +1849,14 @@ void sdiffstoreCommand(client *c) {
 void sscanCommand(client *c) {
     kvobj *set;
     unsigned long long cursor;
+    size_t oldsize = 0;
 
     if (parseScanCursorOrReply(c,c->argv[2],&cursor) == C_ERR) return;
     if ((set = lookupKeyReadOrReply(c,c->argv[1],shared.emptyscan)) == NULL ||
         checkType(c,set,OBJ_SET)) return;
+    if (server.memory_tracking_per_slot)
+        oldsize = setTypeAllocSize(set);
     scanGenericCommand(c,set,cursor);
+    if (server.memory_tracking_per_slot)
+        updateSlotAllocSize(c->db, getKeySlot(c->argv[1]->ptr), oldsize, setTypeAllocSize(set));
 }
