@@ -322,7 +322,7 @@ start_server {tags {"repl external:skip"}} {
 
 foreach mdl {no yes} rdbchannel {no yes} {
     foreach sdl {disabled swapdb} {
-        start_server {tags {"repl external:skip"} overrides {save {}}} {
+        start_server {tags {"repl external:skip debug_defrag:skip"} overrides {save {}}} {
             set master [srv 0 client]
             $master config set repl-diskless-sync $mdl
             $master config set repl-diskless-sync-delay 5
@@ -1505,7 +1505,7 @@ foreach disklessload {disabled on-empty-db} {
                 $replica config set repl-diskless-load $disklessload
 
                 # Populate replica with many keys, master with a few keys.
-                $replica debug populate 2000000
+                $replica debug populate 4000000
                 populate 3 master 10
 
                 # Start the replication process...
@@ -1539,7 +1539,7 @@ foreach disklessload {disabled on-empty-db} {
                 catch {$replica shutdown nosave}
             }
         }
-    } {} {repl external:skip}
+    } {} {repl external:skip debug_defrag:skip}
 }
 
 start_server {tags {"repl external:skip"} overrides {save {}}} {
@@ -1726,10 +1726,9 @@ start_server {tags {"repl external:skip"}} {
             assert {$link_down_since > 0}
             assert {$total_disconnect_time > $link_down_since}
 
-            #  total_disconnect_time did not change after reconnect to real master
+            #  total_disconnect_time_reconnect can be up to 5 seconds more than total_disconnect_time due to reconnection time
             set total_disconnect_time_reconnect [status $slave total_disconnect_time_sec]
-            assert {$total_disconnect_time == $total_disconnect_time_reconnect}
-
+            assert {$total_disconnect_time_reconnect >= $total_disconnect_time && $total_disconnect_time_reconnect <= $total_disconnect_time + 5}
         }
     }
 }
