@@ -525,12 +525,15 @@ int test_malloc_api(RedisModuleCtx *ctx, RedisModuleString **argv, int argc) {
 }
 
 int test_keyslot(RedisModuleCtx *ctx, RedisModuleString **argv, int argc) {
-    /* Static check of the ClusterKeySlot + ClusterCanonicalKeyNameInSlot
+    /* Static check of the ClusterKeySlot + ClusterKeySlotC + ClusterCanonicalKeyNameInSlot
      * round-trip for all slots. */
     for (unsigned int slot = 0; slot < 16384; slot++) {
         const char *tag = RedisModule_ClusterCanonicalKeyNameInSlot(slot);
         RedisModuleString *key = RedisModule_CreateStringPrintf(ctx, "x{%s}y", tag);
         assert(slot == RedisModule_ClusterKeySlot(key));
+        size_t len;
+        const char *key_c = RedisModule_StringPtrLen(key, &len);
+        assert(slot == RedisModule_ClusterKeySlotC(key_c, len));
         RedisModule_FreeString(ctx, key);
     }
     if (argc != 2){
