@@ -3,9 +3,15 @@
 # Copyright (C) 2014-Present, Redis Ltd.
 # All Rights reserved.
 #
+# Copyright (c) 2024-present, Valkey contributors.
+# All rights reserved.
+#
 # Licensed under your choice of (a) the Redis Source Available License 2.0
 # (RSALv2); or (b) the Server Side Public License v1 (SSPLv1); or (c) the
 # GNU Affero General Public License v3 (AGPLv3).
+#
+# Portions of this file are available under BSD3 terms; see REDISCONTRIBUTIONS for more information.
+#
 
 package require Tcl 8.5
 
@@ -24,6 +30,7 @@ set ::all_tests []
 set test_dirs {
     unit
     unit/type
+    unit/moduleapi
     unit/cluster
     integration
 }
@@ -82,6 +89,7 @@ set ::ignoredigest 0
 set ::large_memory 0
 set ::log_req_res 0
 set ::force_resp3 0
+set ::debug_defrag 0
 
 # Set to 1 when we are running in client mode. The Redis test uses a
 # server-client model to run tests simultaneously. The server instance
@@ -237,6 +245,10 @@ proc s {args} {
     status [srv $level "client"] [lindex $args 0]
 }
 
+proc S {index field} {
+    getInfoProperty [R $index info] $field
+}
+
 # Get the specified field from the givens instances cluster info output.
 proc CI {index field} {
     getInfoProperty [R $index cluster info] $field
@@ -576,6 +588,7 @@ proc print_help_screen {} {
         "--ignore-encoding  Don't validate object encoding."
         "--ignore-digest    Don't use debug digest validations."
         "--large-memory     Run tests using over 100mb."
+        "--debug-defrag     Indicate the test is running against server compiled with DEBUG_DEFRAG option"
         "--help             Print this help screen."
     } "\n"]
 }
@@ -707,6 +720,8 @@ for {set j 0} {$j < [llength $argv]} {incr j} {
         set ::ignoreencoding 1
     } elseif {$opt eq {--ignore-digest}} {
         set ::ignoredigest 1
+    } elseif {$opt eq {--debug-defrag}} {
+        set ::debug_defrag 1
     } elseif {$opt eq {--help}} {
         print_help_screen
         exit 0
