@@ -786,7 +786,7 @@ start_cluster 1 0 {tags {external:skip cluster}} {
 
 start_cluster 1 0 {tags {external:skip cluster} overrides {cluster-slot-stats-enabled yes}} {
 
-    set metrics [list "key-count" "cpu-usec" "network-bytes-in" "network-bytes-out"]
+    set metrics [list "key-count" "memory-bytes" "cpu-usec" "network-bytes-in" "network-bytes-out"]
 
     # SET keys for target hashslots, to encourage ordering.
     set hash_tags [list 0 1 2 3 4]
@@ -986,3 +986,125 @@ start_cluster 1 1 {tags {external:skip cluster} overrides {cluster-slot-stats-en
     R 0 CONFIG RESETSTAT
     R 1 CONFIG RESETSTAT
 }
+
+start_cluster 2 2 {tags {external:skip cluster} overrides {cluster-slot-stats-enabled yes}} {
+    test "CLUSTER SLOT-STATS reset upon atomic slot migration" {
+        # key on slot-0
+        set key0 "{06S}mykey0"
+        set key0_slot [R 0 CLUSTER KEYSLOT $key0]
+        R 0 SET $key0 VALUE
+
+        # Migrate slot-0 to node-1
+        R 1 CLUSTER MIGRATION IMPORT 0 0
+        wait_for_condition 1000 10 {
+            [CI 0 cluster_slot_migration_active_tasks] == 0 &&
+            [CI 1 cluster_slot_migration_active_tasks] == 0
+        } else {
+            fail "ASM tasks did not complete"
+        }
+
+        set expected_slot_stats [
+            dict create \
+                $key0_slot [ \
+                    dict create key-count 1 \
+                    dict create cpu-usec 0 \
+                    dict create network-bytes-in 0 \
+                    dict create network-bytes-out 0 \
+                ]
+        ]
+        set metrics_to_assert [list key-count cpu-usec network-bytes-in network-bytes-out]
+
+        # Verify metrics are reset except key-count
+        set slot_stats [R 1 CLUSTER SLOT-STATS SLOTSRANGE 0 0]
+        assert_empty_slot_stats_with_exception $slot_stats $expected_slot_stats $metrics_to_assert
+
+        # Migrate slot-0 back to node-0
+        R 0 CLUSTER MIGRATION IMPORT 0 0
+        wait_for_condition 1000 10 {
+            [CI 0 cluster_slot_migration_active_tasks] == 0 &&
+            [CI 1 cluster_slot_migration_active_tasks] == 0
+        } else {
+            fail "ASM tasks did not complete"
+        }
+
+        # Verify metrics are reset except key-count
+        set slot_stats [R 0 CLUSTER SLOT-STATS SLOTSRANGE 0 0]
+        assert_empty_slot_stats_with_exception $slot_stats $expected_slot_stats $metrics_to_assert
+    }
+}
+
+# -----------------------------------------------------------------------------
+# Test cases for CLUSTER SLOT-STATS memory-bytes field presence.
+# -----------------------------------------------------------------------------
+
+start_cluster 1 0 {tags {external:skip cluster} overrides {cluster-slot-stats-enabled yes}} {
+    # Define shared variables.
+    set key "FOO"
+    set key_slot [R 0 cluster keyslot $key]
+
+    test "CLUSTER SLOT-STATS memory-bytes field present when cluster-slot-stats-enabled set on startup" {
+        R 0 SET $key VALUE
+        set slot_stats [R 0 CLUSTER SLOT-STATS SLOTSRANGE 0 16383]
+        set slot_stats [convert_array_into_dict $slot_stats]
+
+        # Verify memory-bytes field is present
+        assert {[dict exists $slot_stats $key_slot]}
+        set stats [dict get $slot_stats $key_slot]
+        assert {[dict exists $stats memory-bytes]}
+        assert {[dict get $stats memory-bytes] > 0}
+    }
+
+    test "CLUSTER SLOT-STATS memory-bytes field still present after disabling cluster-slot-stats-enabled" {
+        R 0 CONFIG SET cluster-slot-stats-enabled no
+        set slot_stats [R 0 CLUSTER SLOT-STATS SLOTSRANGE 0 16383]
+        set slot_stats [convert_array_into_dict $slot_stats]
+
+        # Verify memory-bytes field is still present even after disabling config
+        assert {[dict exists $slot_stats $key_slot]}
+        set stats [dict get $slot_stats $key_slot]
+        assert {[dict exists $stats memory-bytes]}
+        assert {[dict get $stats memory-bytes] > 0}
+
+        # Verify other stats fields are not present
+        assert {![dict exists $stats cpu-usec]}
+        assert {![dict exists $stats network-bytes-in]}
+        assert {![dict exists $stats network-bytes-out]}
+    }
+}
+
+start_cluster 1 0 {tags {external:skip cluster} overrides {cluster-slot-stats-enabled no}} {
+    # Define shared variables.
+    set key "FOO"
+    set key_slot [R 0 cluster keyslot $key]
+
+    test "CLUSTER SLOT-STATS memory-bytes field not present when cluster-slot-stats-enabled not set on startup" {
+        R 0 SET $key VALUE
+        set slot_stats [R 0 CLUSTER SLOT-STATS SLOTSRANGE 0 16383]
+        set slot_stats [convert_array_into_dict $slot_stats]
+
+        # Verify memory-bytes field is not present
+        assert {[dict exists $slot_stats $key_slot]}
+        set stats [dict get $slot_stats $key_slot]
+        assert {![dict exists $stats memory-bytes]}
+
+        # Only key-count should be present
+        assert {[dict exists $stats key-count]}
+        assert {[dict get $stats key-count] == 1}
+    }
+
+    test "CLUSTER SLOT-STATS memory-bytes field not present after enabling cluster-slot-stats-enabled via CONFIG SET" {
+        R 0 CONFIG SET cluster-slot-stats-enabled yes
+        set slot_stats [R 0 CLUSTER SLOT-STATS SLOTSRANGE 0 16383]
+        set slot_stats [convert_array_into_dict $slot_stats]
+
+        # Verify memory-bytes field is still not present
+        assert {[dict exists $slot_stats $key_slot]}
+        set stats [dict get $slot_stats $key_slot]
+        assert {![dict exists $stats memory-bytes]}
+
+        # Other stats fields should now be present
+        assert {[dict exists $stats cpu-usec]}
+        assert {[dict exists $stats network-bytes-in]}
+        assert {[dict exists $stats network-bytes-out]}
+    }
+}
