@@ -10,7 +10,7 @@ set testmodule_blockedclient [file normalize tests/modules/blockedclient.so]
 set testmodule [file normalize tests/modules/blockonkeys.so]
 
 set modules [list loadmodule $testmodule loadmodule $testmodule_nokey loadmodule $testmodule_blockedclient]
-start_cluster 3 0 [list config_lines $modules] {
+start_cluster 3 0 [list tags {external:skip cluster modules} config_lines $modules] {
 
     set node1 [srv 0 client]
     set node2 [srv -1 client]
@@ -166,7 +166,7 @@ start_cluster 3 0 [list config_lines $modules] {
 set testmodule_keyspace_events [file normalize tests/modules/keyspace_events.so]
 set testmodule_postnotifications "[file normalize tests/modules/postnotifications.so] with_key_events"
 set modules [list loadmodule $testmodule_keyspace_events loadmodule $testmodule_postnotifications]
-start_cluster 2 2 [list config_lines $modules] {
+start_cluster 2 2 [list tags {external:skip cluster modules} config_lines $modules] {
 
     set master1 [srv 0 client]
     set master2 [srv -1 client]
@@ -213,7 +213,7 @@ start_cluster 2 2 [list config_lines $modules] {
 
 set testmodule [file normalize tests/modules/basics.so]
 set modules [list loadmodule $testmodule]
-start_cluster 3 0 [list config_lines $modules] {
+start_cluster 3 0 [list tags {external:skip cluster modules} config_lines $modules] {
     set node1 [srv 0 client]
     set node2 [srv -1 client]
     set node3 [srv -2 client]
