@@ -1,6 +1,6 @@
 set testmodule [file normalize tests/modules/commandfilter.so]
 
-start_server {tags {"modules"}} {
+start_server {tags {"modules external:skip"}} {
     r module load $testmodule log-key 0
 
     test {Retain a command filter argument} {
@@ -95,7 +95,7 @@ start_server {tags {"modules"}} {
     test "Unload the module - commandfilter" {
         assert_equal {OK} [r module unload commandfilter]
     }
-} 
+}
 
 test {RM_CommandFilterArgInsert and script argv caching} {
     # coverage for scripts calling commands that expand the argv array
@@ -103,7 +103,7 @@ test {RM_CommandFilterArgInsert and script argv caching} {
     # this test needs a fresh server so that lua_argv_size is 0.
     # glibc realloc can return the same pointer even when the size changes
     # still this test isn't able to trigger the issue, but we keep it anyway.
-    start_server {tags {"modules"}} {
+    start_server {tags {"modules external:skip"}} {
         r module load $testmodule log-key 0
         r del mylist
         # command with 6 args
@@ -116,3 +116,60 @@ test {RM_CommandFilterArgInsert and script argv caching} {
     }
 }
 
+# previously, there was a bug that command filters would be rerun (which would cause args to swap back)
+# this test is meant to protect against that bug
+test {Blocking Commands don't run through command filter when reprocessed} {
+    start_server {tags {"modules external:skip"}} {
+        r module load $testmodule log-key 0
+
+        r del list1{t}
+        r del list2{t}
+
+        r lpush list2{t} a b c d e
+
+        set rd [redis_deferring_client]
+        # we're asking to pop from the left, but the command filter swaps the two arguments,
+        # if it didn't swap it, we would end up with e d c b a 5 (5 being the left most of the following lpush)
+        # but since we swap the arguments, we end up with 1 e d c b a (1 being the right most of it).
+        # if the command filter would run again on unblock, they would be swapped back.
+        $rd blmove list1{t} list2{t} left right 0
+        wait_for_blocked_client
+        r lpush list1{t} 1 2 3 4 5
+        # validate that we moved the correct element with the swapped args
+        assert_equal [$rd read] 1
+        # validate that we moved the correct elements to the correct side of the list
+        assert_equal [r lpop list2{t}] 1
+
+        $rd close
+    }
+}
+
+test {Filtering based on client id} {
+    start_server {tags {"modules external:skip"}} {
+        r module load $testmodule log-key 0
+
+        set rr [redis_client]
+        set cid [$rr client id]
+        r unfilter_clientid $cid
+
+        r rpush mylist elem1 @replaceme elem2
+        assert_equal [r lrange mylist 0 -1] {elem1 --replaced-- elem2}
+
+        r del mylist
+
+        assert_equal [$rr rpush mylist elem1 @replaceme elem2] 3
+        assert_equal [r lrange mylist 0 -1] {elem1 @replaceme elem2}
+
+        $rr close
+    }
+}
+
+start_server {tags {"external:skip"}} {
+    test {OnLoad failure will handle un-registration} {
+        catch {r module load $testmodule log-key 0 noload}
+        r set mykey @log
+        assert_equal [r lrange log-key 0 -1] {}
+        r rpush mylist elem1 @delme elem2
+        assert_equal [r lrange mylist 0 -1] {elem1 @delme elem2}
+    }
+}
