@@ -1,6 +1,6 @@
 set testmodule [file normalize tests/modules/keyspace_events.so]
 
-tags "modules" {
+tags "modules external:skip" {
     start_server [list overrides [list loadmodule "$testmodule"]] {
 
         # avoid using shared integers, to increase the chance of detection heap issues
