@@ -1,6 +1,6 @@
 set testmodule [file normalize tests/modules/rdbloadsave.so]
 
-start_server {tags {"modules"}} {
+start_server {tags {"modules external:skip debug_defrag:skip"}} {
     r module load $testmodule
 
     test "Module rdbloadsave sanity" {
@@ -147,11 +147,11 @@ start_server {tags {"modules"}} {
 
     tags {repl} {
         test {Module rdbloadsave on master and replica} {
-            start_server [list overrides [list loadmodule "$testmodule"]] {
+            start_server [list overrides [list loadmodule "$testmodule"] tags {"external:skip"}] {
                 set replica [srv 0 client]
                 set replica_host [srv 0 host]
                 set replica_port [srv 0 port]
-                start_server [list overrides [list loadmodule "$testmodule"]] {
+                start_server [list overrides [list loadmodule "$testmodule"] tags {"external:skip"}] {
                     set master [srv 0 client]
                     set master_host [srv 0 host]
                     set master_port [srv 0 port]
