@@ -940,6 +940,18 @@ start_server {tags {"hash"}} {
         set _ $k
     } {ZIP_INT_8B 127 ZIP_INT_16B 32767 ZIP_INT_32B 2147483647 ZIP_INT_64B 9223372036854775808 ZIP_INT_IMM_MIN 0 ZIP_INT_IMM_MAX 12}
 
+    test {KEYS command return expired keys when allow_access_expired is 1} {
+        r flushall
+        r debug set-allow-access-expired 1
+        r debug set-active-expire 0
+        r set key1 value1
+        r pexpire key1 1
+        after 2
+        assert_equal {key1} [r keys *]
+        r debug set-allow-access-expired 0
+        r debug set-active-expire 1
+    } {OK} {needs:debug}
+
     # On some platforms strtold("+inf") with valgrind returns a non-inf result
     if {!$::valgrind} {
         test {HINCRBYFLOAT does not allow NaN or Infinity} {
